<?php

namespace Zen\Paymentpage\Controller\Payment;

use Exception;
use Magento\Framework\App\Action\Action;
use Magento\Framework\App\Action\Context;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\ResponseInterface;
use Magento\Framework\Controller\ResultInterface;
use Magento\Sales\Api\InvoiceRepositoryInterface;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\Sales\Api\OrderStatusHistoryRepositoryInterface;
use Magento\Sales\Model\Order;
use Magento\Sales\Model\Order\CreditmemoFactory;
use Magento\Sales\Model\Order\Email\Sender\InvoiceSender;
use Magento\Sales\Model\Order\Invoice;
use Magento\Sales\Model\Order\Payment\Transaction;
use Magento\Sales\Model\Order\Payment\Transaction\BuilderInterface;
use Magento\Sales\Model\OrderRepository;
use Magento\Sales\Model\Service\CreditmemoService;
use Magento\Sales\Model\Service\InvoiceService;
use Zen\Payment\Util;

/**
 * Class Notification
 *
 * @package Zen\Paymentpage\Controller\Payment
 */
class Notification extends Action
{

	/**
	 * @var Order
	 */
	protected $order;

	/**
	 * @var ScopeConfigInterface
	 */
	protected $scopeConfig;

	/**
	 * @var BuilderInterface
	 */
	protected $builderInterface;

	/**
	 * @var OrderRepository
	 */
	protected $orderRepository;

	/**
	 * @var InvoiceService
	 */
	protected $invoiceService;

	/**
	 * @var \Magento\Framework\DB\Transaction
	 */
	protected $dbTransaction;

	/**
	 * @var Magento\Sales\Model\Order\Email\Sender\InvoiceSender
	 */
	protected $invoiceSender;

	/**
	 * @var Magento\Sales\Model\Order\CreditmemoFactory
	 */
	protected $creditmemoFactory;

	/**
	 * @var Magento\Sales\Model\Service\CreditmemoService
	 */
	protected $creditmemoService;

	/**
	 * @var Magento\Sales\Api\InvoiceRepositoryInterface
	 */
	protected $invoiceRepository;

	/**
	 * @var OrderStatusHistoryRepositoryInterface
	 */
	protected $orderStatusRepository;

	/**
	 * ReturnUrl constructor.
	 *
	 * @param Context                               $context
	 * @param Order                                 $order
	 * @param ScopeConfigInterface                  $scopeConfig
	 * @param BuilderInterface                      $builderInterface
	 * @param OrderRepositoryInterface              $orderRepository
	 * @param InvoiceService                        $invoiceService
	 * @param \Magento\Framework\DB\Transaction     $dbTransaction
	 * @param InvoiceSender                         $invoiceSender
	 * @param CreditmemoFactory                     $creditmemoFactory
	 * @param CreditmemoService                     $creditmemoService
	 * @param InvoiceRepositoryInterface            $invoiceRepository
	 * @param OrderStatusHistoryRepositoryInterface $orderStatusRepository
	 */
	public function __construct(
		Context $context,
		Order $order,
		ScopeConfigInterface $scopeConfig,
		BuilderInterface $builderInterface,
		OrderRepositoryInterface $orderRepository,
		InvoiceService $invoiceService,
		\Magento\Framework\DB\Transaction $dbTransaction,
		InvoiceSender $invoiceSender,
		CreditmemoFactory $creditmemoFactory,
		CreditmemoService $creditmemoService,
		InvoiceRepositoryInterface $invoiceRepository,
		OrderStatusHistoryRepositoryInterface $orderStatusRepository
	) {
		parent::__construct($context);
		$this->order = $order;
		$this->scopeConfig = $scopeConfig;
		$this->builderInterface = $builderInterface;
		$this->orderRepository = $orderRepository;
		$this->invoiceService = $invoiceService;
		$this->dbTransaction = $dbTransaction;
		$this->invoiceSender = $invoiceSender;
		$this->creditmemoFactory = $creditmemoFactory;
		$this->creditmemoService = $creditmemoService;
		$this->invoiceRepository = $invoiceRepository;
		$this->orderStatusRepository = $orderStatusRepository;
	}

	/**
	 * @return ResponseInterface|ResultInterface|void
	 * @throws Exception
	 */
	public function execute()
	{

		$notification = new \Zen\Payment\Notification(
			$this->scopeConfig->getValue('payment/zen_paymentpage/ipn_secret')
		);

		// it can be order data or notification code - depends on verification notification
		$resultCheckRequestNotification = $notification->checkRequest();

		if(is_int($resultCheckRequestNotification)) {

			echo \Zen\Payment\Notification::formatResponse(\Zen\Payment\Notification::NS_ERROR);
			exit();
		}

		if($resultCheckRequestNotification['type'] === \Zen\Payment\Notification::TRT_REFUND) {

			if(!($order = $this->order->loadByIncrementId(Util::extractMerchantTransactionId($resultCheckRequestNotification['merchantTransactionId'])))) {
				echo \Zen\Payment\Notification::formatResponse(\Zen\Payment\Notification::NS_ERROR);
				exit();
			}

			$currentOrderStatus = $order->getState();

			if($currentOrderStatus === Order::STATE_CLOSED) {

				echo \Zen\Payment\Notification::formatResponse(\Zen\Payment\Notification::NS_ERROR);
				exit();
			}

			try {
				$creditmemo = $this->creditmemoFactory->createByOrder($order);
				$creditmemo->setGrandTotal($resultCheckRequestNotification['amount']);
				$this->creditmemoService->refund($creditmemo);
				$this->creditmemoService->notify($creditmemo->getId());

				$this->orderStatusRepository->save(__('Refund for amount:')
					. ' '
					. $resultCheckRequestNotification['amount']
					. ' '
					. $order->getOrderCurrency()->getCode()
					. ' '
					. __('with Zen UUID:')
					. ' '
					. $resultCheckRequestNotification['transactionId']
					. ' '
					. __('has been correctly processed')
					. '. '
					. __('Below refund "offline" is the same.'));
			} catch(\Exception $e) {

				echo \Zen\Payment\Notification::formatResponse(\Zen\Payment\Notification::NS_ERROR);
				exit();
			}

			echo \Zen\Payment\Notification::formatResponse(\Zen\Payment\Notification::NS_OK);
			exit();
		}

		if(!($order = $this->order->loadByIncrementId($resultCheckRequestNotification['merchantTransactionId']))) {

			echo \Zen\Payment\Notification::formatResponse(\Zen\Payment\Notification::NS_ERROR);
			exit();
		}

		if(!\Zen\Payment\Notification::checkRequestAmount($resultCheckRequestNotification, $order->getGrandTotal(), $order->getOrderCurrencyCode())) {

			echo \Zen\Payment\Notification::formatResponse(\Zen\Payment\Notification::NS_ERROR);
			exit();
		}

		$currentOrderStatus = $order->getState();

		if($currentOrderStatus === Order::STATE_PROCESSING || $currentOrderStatus === Order::STATE_CLOSED) {

			echo \Zen\Payment\Notification::formatResponse(\Zen\Payment\Notification::NS_OK);
			exit();
		}

		switch($resultCheckRequestNotification['status']) {
			case \Zen\Payment\Notification::STATUS_ACCEPTED:

				$order->setTotalPaid($resultCheckRequestNotification['amount']);
				$this->createTransaction($order, $resultCheckRequestNotification);
				$this->createInvoice($order, $resultCheckRequestNotification['transactionId']);
				$order->save();
				break;
			case \Zen\Payment\Notification::STATUS_REJECTED:
				$orderState = Order::STATE_PAYMENT_REVIEW;
				$order->setState($orderState)->setStatus($orderState);
				$order->save();
				$order->addStatusToHistory($orderState, __('The payment has been rejected.'));
				break;
			default:
				break;
		}

		echo \Zen\Payment\Notification::formatResponse(\Zen\Payment\Notification::NS_OK);
		exit();
	}

	/**
	 * @param Order $order
	 * @param array $payloadDecoded
	 *
	 * @return bool
	 */
	private function createTransaction($order, $payloadDecoded)
	{
		//get payment object from order object
		//get the object of builder class
		if(
			!($payment = $order->getPayment())
			|| !($trans = $this->builderInterface)
		) {
			return false;
		}

		$transaction = $trans->setPayment($payment)
			->setOrder($order)
			->setTransactionId($payloadDecoded['transactionId'])
			->setFailSafe(true)
			//build method creates the transaction and returns the object
			->build(Transaction::TYPE_CAPTURE);
		if(!$transaction) {
			return false;
		}

		$payment->addTransactionCommentsToOrder(
			$transaction,
			__('Paid amount: %1.', Util::convertAmountToMain($payloadDecoded['amount']))
		);

		$payment->setTransactionId($payloadDecoded['transactionId']);

		$payment->setParentTransactionId(null);

		try {
			$payment->save();
			$transaction->save();

			return true;
		} catch(Exception $e) {
			return false;
		}
	}

	/**
	 * @param Order  $order
	 * @param string $transactionId
	 *
	 * @return bool
	 */
	private function createInvoice($order, $transactionId)
	{

		if(!$order->canInvoice()) {

			return false;
		}

		$invoice = $this->invoiceService->prepareInvoice($order);
		$invoice->setRequestedCaptureCase(Invoice::CAPTURE_ONLINE);
		$invoice->setTransactionId($transactionId);
		$invoice->register();
		$invoice->pay();
		$invoice->save();

		if($invoice && !$invoice->getEmailSent()) {

			$this->invoiceSender->send($invoice);

			$order->addRelatedObject($invoice);
			$order->addStatusHistoryComment(__('You notified customer about invoice #%1.', $invoice->getIncrementId()))
				->setIsCustomerNotified(true)
				->save();

			return true;
		}

		return false;
	}
}
